#include "Eigen/Dense"
#include <iostream>
#include <cmath>

template<typename T>
T f(const T &p, const T &x) { return pow(p*x,2); }

template<typename T>
T dfdp(const T &p, const T &x) { return 2*p*pow(x,2); }

template<typename T>
T ddfdpp(const T&, const T &x) { return 2*pow(x,2); }

template<typename T, int M>
T E(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T o=0;
  for (int i=0;i<M;i++) o+=pow(f(p,x(i))-y(i),2);
  return o;
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T r=0;
  for (int i=0;i<M;i++) r+=(f(p,x(i))-y(i))*dfdp(p,x(i));
  return 2*r;
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T r=0;
  for (int i=0;i<M;i++)
    r+=dfdp(p,x(i))*dfdp(p,x(i))+(f(p,x(i))-y(i))*ddfdpp(p,x(i));
  return 2*r;
}

template<typename T, int M>
void Bisection(T& p_l, T& p_u, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) {
  while (fabs(p_l-p_u)>eps) {
    std::cout << ".";
    T p_m=(p_u+p_l)/2; 
    T r_m=dEdp(p_m,x,y); 
    if (r_m>0) p_u=p_m;
    else if(r_m<0) p_l=p_m;
    else { p_l=p_u=p_m; break; }
  }
  std::cout << std::endl;
}

int main() {
  const int M=3;
  using T=double;
  using VT=Eigen::Matrix<T,M,1>;
  VT x=VT::Random(); x=x.cwiseProduct(x);
  std::cout << "x=^T" << x.transpose() << std::endl;
  VT y=VT::Random(); 
  std::cout << "y^T=" << y.transpose() << std::endl;
  T p=1, p_l=p-10, p_u=p+10;
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  Bisection(p_l,p_u,x,y,1e-7);
  std::cout << "p=" << p_l << "; E=" << E(p_l,x,y) << "; dEdp=" 
	    << dEdp(p_l,x,y) << "; ddEdpp=" << ddEdpp(p_l,x,y) << std::endl;
  return 0;
}
