#include "Eigen/Dense"
#include <iostream>
#include <cmath>

template<typename T>
T f(const T &p, const T &x) { return pow(p*x,2); }

template<typename T>
T dfdp(const T &p, const T &x) { return 2*p*pow(x,2); }

template<typename T>
T ddfdpp(const T&, const T &x) { return 2*pow(x,2); }

template<typename T, int M>
T E(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T o=0;
  for (int i=0;i<M;i++) o+=pow(f(p,x(i))-y(i),2);
  return o;
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T r=0;
  for (int i=0;i<M;i++) r+=(f(p,x(i))-y(i))*dfdp(p,x(i));
  return 2*r;
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T r=0;
  for (int i=0;i<M;i++)
    r+=dfdp(p,x(i))*dfdp(p,x(i))+(f(p,x(i))-y(i))*ddfdpp(p,x(i));
  return 2*r;
}

template<typename T, int M>
void SteepestDescent(T& p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) {
  T g=dEdp(p,x,y), e=E(p,x,y),e_prev;
  while (fabs(g)>eps) {
    std::cout << ".";
    e_prev=e;
    T alpha=2.;
    while (e_prev<=e&&alpha>eps) {
      T p_trial=p; alpha/=2;
      p_trial-=alpha*g; e=E(p_trial,x,y);
    }
    p-=alpha*g; g=dEdp(p,x,y);
  }
  std::cout << std::endl;
}

int main() {
  const int M=3;
  using T=double;
  using VT=Eigen::Matrix<T,M,1>;
  VT x=VT::Random(); x=x.cwiseProduct(x);
  std::cout << "x=^T" << x.transpose() << std::endl;
  VT y=VT::Random(); 
  std::cout << "y^T=" << y.transpose() << std::endl;
  T p=1;
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  SteepestDescent(p,x,y,1e-7);
  std::cout << "p=" << p << "; E=" << E(p,x,y) << "; dEdp=" << dEdp(p,x,y) 
	    << "; ddEdpp=" << ddEdpp(p,x,y) << std::endl;
  return 0;
}
